<?php

namespace Drupal\search_api_exactmatch_boost\Plugin\search_api\processor;

use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\PluginFormInterface;
use Drupal\search_api\Plugin\PluginFormTrait;
use Drupal\search_api\Processor\FieldsProcessorPluginBase;
use Drupal\search_api\Query\ResultSetInterface;

/**
 * Boosts a result if it matches exactly the search key
 *
 * This processor will only run if the first sort of the query is relevance,
 * it will resort the results array after adding boost for exact matches,
 * it uses the original query keys for comparing, so manipulation on the
 * query keys done through proccessing elsewhere will have no effect here
 *
 * @SearchApiProcessor(
 *   id = "exactmatchboost",
 *   label = @Translation("Exact match boosting"),
 *   description = @Translation("Boost an exact match of search key in a field. Will only run if relevance is highest sort order, mainly intended for plain text fields like title and partial matching."),
 *   stages = {
 *     "postprocess_query" = 0,
 *   }
 * )
 */
class ExactMatchBoost extends FieldsProcessorPluginBase implements PluginFormInterface {

  use PluginFormTrait;

  /**
   * {@inheritdoc}
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    $configuration = parent::defaultConfiguration();
    $configuration += [
      'exactmatch' => [
        'scoreadd' => 0,
        'ignorecase' => 1,
      ],
    ];
    return $configuration;
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);
    $form['exactmatch']['scoreadd'] = [
      '#type' => 'number',
      '#title' => $this->t('Score to add to relevance if search key matches exactly'),
      '#default_value' => $this->configuration['exactmatch']['scoreadd']
    ];
    $form['exactmatch']['ignorecase'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Ignore case'),
      '#default_value' => isset($this->configuration['exactmatch']['ignorecase']) ? $this->configuration['exactmatch']['ignorecase'] : 1
    ];
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    $values = $form_state->getValues();
    $configuration = $this->getConfiguration();
    if(!empty(array_diff($values['fields'], $configuration['fields']))) {
      $form_state->set('processors_changed', TRUE);
    }
    $this->setConfiguration($values);
  }

  /**
   * {@inheritdoc}
   */
  public function postprocessSearchResults(ResultSetInterface $results) {
    $configuration = $this->configuration;
    $query = $results->getQuery();
    $searchkeys = $query->getOriginalKeys();
    if ($this->configuration['exactmatch']['ignorecase']) {
      $searchkeys = mb_strtolower($searchkeys);
    }
    $sorts = $query->getSorts();
    $first = array_splice($sorts, 0, 1);
    if (key($first) == 'search_api_relevance') {
      $direction = $first['search_api_relevance'];
      $resultsreorder = $results->getResultItems();
      $resort = FALSE;
      foreach($configuration['fields'] as $field) {
        foreach ($resultsreorder as $key => $res) {
          if ($res->getField($field)) {
            $values = $res->getField($field)->getValues();
            foreach ($values as $val) {
              $comparevalue = $val->getText();
              if ($this->configuration['exactmatch']['ignorecase']) {
                $comparevalue = mb_strtolower($comparevalue);
              }
              if ($comparevalue == $searchkeys) {
                $resultsreorder[$key]->setScore($resultsreorder[$key]->getScore() + $this->configuration['exactmatch']['scoreadd']);
                $resort = TRUE;
              }
            }
          }
        }
      }
      if ($resort) {
        uasort($resultsreorder, [
          'Drupal\search_api_exactmatch_boost\Plugin\search_api\processor\ExactMatchBoost',
          'newScoreSort'
        ]);
        if ($direction == 'ASC') {
          $resultsreorder = array_reverse($resultsreorder);
        }
        $results->setResultItems($resultsreorder);
      }
    }
    parent::postprocessSearchResults($results);
  }

  /**
   * callback to resort array of result items based on score (relevance)
   *
   * @param $a - result item
   * @param $b - result item
   *
   * @return int
   */
  private static function newScoreSort($a, $b) {
    if($a->getScore() == $b->getScore()) {
      return 0;
    }
    return ($a->getScore() < $b->getScore()) ? 1 : -1;
  }

}
